indexing
	description: "Objects that instrument Eiffel code with computational state tracking instructions"
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_INSTRUMENTER

inherit
	CI_PROGRAM_POINT_FINDER
		rename
			error_handler as processor_error_handler
		export {NONE}
			processor_error_handler
		redefine
			make,
			process_class,
--			process_deferred_procedure,
--			process_deferred_function,
			process_do_procedure,
			process_do_function,
--			process_external_procedure,
--			process_external_function,
			process_once_procedure,
			process_once_function,
			process_program_point
		end

	CI_EXCLUDED_ITEMS

	CI_ERROR_HANDLER_CLIENT

	CI_IDENTITIES_NAMES_CLIENT

	CI_RUNTIME_OUTPUT_METACLIENT
		export {NONE}
			all
		end

create
	make

feature -- Initialization
	make (a_universe: like universe; a_printer: like printer) is
			-- Create instrumenter that works within `a_universe' with program point information printer `a_printer'
		do
			Precursor (a_universe, a_printer)
			create error_handler.make_null
			create state_printing_functions.make
			create identities_metaclient.make (universe)
			create runtime_output_metaclient
		end

feature -- Processing
	process_system is
			-- Instrument classes
		local
			ppt: CI_PROGRAM_POINT
			cursor: DS_HASH_SET_CURSOR [ET_CLASS_NAME]
		do
			-- Pass 1: gather information about program points, create instrumentation code
			from
				cursor := Settings.classes_to_instrument.new_cursor
				cursor.start
			until
				cursor.after
			loop
				universe.classes.item (cursor.item).process (Current)
				cursor.forth
			end
			-- Pass 2: insert instrumentation code into classes
			from
				program_points.start
			until
				program_points.after
			loop
				ppt := program_points.item_for_iteration
				ppt.instrument
				program_points.forth
			end
			-- Pass 3: perform additional code changes
			from
				universe.classes.start
			until
				universe.classes.after
			loop
				current_class := universe.classes.item_for_iteration
				add_once_state_recording
				universe.classes.forth
			end
			program_points.wipe_out
		end

feature {ET_AST_NODE} -- Processing
	process_class (a_class: ET_CLASS) is
			-- Process child nodes, create and process object program point for `a_class'
			-- In all `process_<some_node>' procedures child nodes are processed first,
			-- because for routine body program points we want to get positions in original (not instrumented) code
		local
			feature_clauses: ET_FEATURE_CLAUSE_LIST
			state_printing_feature_clause: ET_FEATURE_CLAUSE
			monitoring_control_type: ET_CLASS_TYPE
			monitoring_control_parent: ET_PARENT
		do
			error_handler.report_instrumenting_class (a_class.name)
			state_printing_functions.wipe_out
			Precursor (a_class)

			-- Add query that is target for identities calls
			if a_class.feature_clauses /= Void then
				a_class.feature_clauses.force_first (identities_metaclient.feature_clause)
			else
				create feature_clauses.make_with_capacity (1)
				feature_clauses.put_first (identities_metaclient.feature_clause)
				a_class.set_feature_clauses (feature_clauses)
			end
			a_class.queries.force_first (identities_metaclient.query (a_class))

			create state_printing_feature_clause.make (clients_any)
			add_comment ("Instrumentation", state_printing_feature_clause)
			a_class.feature_clauses.force_first (state_printing_feature_clause)
			from
				state_printing_functions.start
			until
				state_printing_functions.after
			loop
				a_class.queries.force_first (state_printing_functions.item_for_iteration)
				a_class.queries.first.set_feature_clause (state_printing_feature_clause)
				state_printing_functions.forth
			end

			runtime_output_metaclient.update_parents (a_class, universe)
			if Settings.monitoring_control_enabled then
				check some_parents_added: a_class.parents /= Void end
				create monitoring_control_type.make (Void, Settings.Monitoring_control_class_name, universe.eiffel_class (Settings.Monitoring_control_class_name))
				create monitoring_control_parent.make (monitoring_control_type, Void, Void, Void, Void, Void)
				a_class.parents.force_first (monitoring_control_parent)
			end

			a_class.set_group (Settings.instrumented_cluster)
		end

--	process_deferred_procedure (a_routine: ET_DEFERRED_PROCEDURE) is
--			-- Process child nodes, create and process entry and exit program points for `a_routine'
--		do
--			error_handler.report_instrumenting_routine (a_routine.name)
--			Precursor (a_routine)
--		end

--	process_deferred_function (a_routine: ET_DEFERRED_FUNCTION) is
--			-- Process child nodes, create and process entry and exit program points for `a_routine'
--		do
--			error_handler.report_instrumenting_routine (a_routine.name)
--			Precursor (a_routine)
--		end

	process_do_procedure (a_routine: ET_DO_PROCEDURE) is
			-- Process child nodes, create and process entry and exit program points for `a_routine'
		do
			if not excluded_features.has (a_routine) then
				error_handler.report_instrumenting_routine (a_routine.name)
				Precursor (a_routine)
			end
		end

	process_do_function (a_routine: ET_DO_FUNCTION) is
			-- Process child nodes, create and process entry and exit program points for `a_routine'
		do
			if not excluded_features.has (a_routine) then
				error_handler.report_instrumenting_routine (a_routine.name)
				Precursor (a_routine)
			end
		end

--	process_external_procedure (a_routine: ET_EXTERNAL_PROCEDURE) is
--			-- Create and process entry and exit program points for `a_routine'
--		do
--			-- TODO: When adding a post, becomes not available for static call			
--			error_handler.report_instrumenting_routine (a_routine.name)
--			Precursor (a_routine)
--		end

--	process_external_function (a_routine: ET_EXTERNAL_FUNCTION) is
--			 -- Create and process entry and exit program points for `a_routine'
--		do
--			-- TODO: When adding a post, becomes not available for static call
--			error_handler.report_instrumenting_routine (a_routine.name)
--			Precursor (a_routine)
--		end

	process_once_procedure (a_routine: ET_ONCE_PROCEDURE) is
			 -- Process child nodes, create and process entry and exit program points for `a_routine'
		do
			if not excluded_features.has (a_routine) then
				error_handler.report_instrumenting_routine (a_routine.name)
				Precursor (a_routine)
			end
		end

	process_once_function (a_routine: ET_ONCE_FUNCTION) is
			-- Process child nodes, create and process entry and exit program points for `a_routine'
		do
			if not a_routine.name.same_feature_name (RO_names.trace_feature_name) and not excluded_features.has (a_routine) then
				error_handler.report_instrumenting_routine (a_routine.name)
				Precursor (a_routine)
			end
		end

feature {NONE} -- Implementation
	state_printing_functions: LINKED_LIST [CI_STATE_PRINTING_FUNCTION]
			-- Functions that print state at program points of interest (should be added to instrumented class)

	identities_metaclient: CI_IDENTITIES_METACLIENT
			-- Code generator that makes instrumented classes clients of 'identities' class

	runtime_output_metaclient: CI_RUNTIME_OUTPUT_METACLIENT
			-- Code generator that makes instrumented classes descendants of 'runtime_output_client' class

	process_program_point (a_ppt: CI_PROGRAM_POINT) is
			-- Process `a_ppt'
		do
			Precursor (a_ppt)
			printer.process (a_ppt)
			state_printing_functions.extend (a_ppt.state_printing_function (printer.state_printing_instructions, printer.extra_locals))
		end

	add_once_state_recording is
			-- Instrument proper once functions with instruction that records they are already called
		local
			i: INTEGER
			has_changed: BOOLEAN
			internal_function: ET_INTERNAL_FUNCTION
		do
			from
				i := current_class.queries.declared_count
			until
				i < 1
			loop
				internal_function ?= current_class.queries.item (i)
				if internal_function /= Void and Settings.is_valid_query_variable (internal_function) and
					Settings.is_watched_once (internal_function) and
					not Identities_names.is_identity_target (internal_function.name, current_class.name) then
					if internal_function.compound /= Void then
						internal_function.compound.force_first (set_called_instruction (current_class.name, internal_function.name))
					else
						internal_function.set_compound (single_instruction (set_called_instruction (current_class.name, internal_function.name)))
					end
					has_changed := True
				end
				i := i - 1
			end
			if has_changed and not Settings.classes_to_instrument.has (current_class.name) then
				runtime_output_metaclient.update_parents (current_class, universe)
				current_class.set_group (Settings.instrumented_cluster)
				Settings.extra_changed_classes.force (current_class.name)
			end
		end
invariant
	identities_metaclient_exists: identities_metaclient /= Void
	runtime_output_metaclient_exists: runtime_output_metaclient /= Void
end
