indexing
	description: "Program points that correspond to loops, but contain only variables that are not modified (i.e. assigned to) inside the loop"
	author: "Flaviu Roman"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_SURROGATE_LOOP

inherit
	CI_LOOP
		rename
			make as make_loop
		export {NONE}
			make_loop
		redefine
			add_variables,
			kind,
			state_printing_function_name,
			unfolded_variables,
			annotate,
			instrument
		end

create
	make

feature -- Initialization
	make (a_list: like modified_variables; a_loop: like loop_point) is
			-- Create a surrogate loop program point that corresponds to `loop_point', where `a_list' is a list of writable entities, modified inside the loop
		require
			a_list_exists: a_list /= Void
			a_loop_exists: a_loop /= Void
		do
			loop_point := a_loop
			modified_variables := a_list
			make_loop (a_loop.loop_instruction, a_loop.position, a_loop.surrounding_routine, a_loop.surrounding_class, a_loop.universe)
		ensure
			modified_variables_set: modified_variables = a_list
			loop_point_set: loop_point = a_loop
		end

feature -- Access
	loop_point: CI_LOOP
			-- Loop program point that corresponds to current surrogate loop

	modified_variables: LINEAR_SUBSET [ET_WRITABLE]
			-- Writable entities modified inside the loop body

	add_variables is
			-- Find variables in scope and store them into `variables'
		do
			from
				loop_point.variables.start
			until
				loop_point.variables.after
			loop
				if not is_modified (loop_point.variables.item) then
					variables.extend (loop_point.variables.item)
				end
				loop_point.variables.forth
			end
		end

	unfolded_variables:	LIST [TREE [CI_VARIABLE]] is
			-- Unfolded variables (in a tree form)
		local
			tree: TWO_WAY_TREE [CI_VARIABLE]
		do
			if internal_unfolded_variables = Void then
				create internal_unfolded_variables.make (variables.count)
				from
					variables.start
				until
					variables.after
				loop
					tree := variables.item.unfolded
					if not is_modified (tree.item) then
						prune_modified_nodes (tree)
						internal_unfolded_variables.extend (tree)
					end
					variables.forth
				end
			end
			Result := internal_unfolded_variables
		end

	kind (a_visitor: CI_PROGRAM_POINT_KIND_FORMAT): STRING is
			-- Program point kind name
		do
			Result := a_visitor.surrogate_loop_kind
		end

feature -- Assertions
	instrument is
			-- Instrument program point by making `checked_assertions' call `state_printing_function'
		do
		end

	annotate is
			-- Do nothing because the point is surrogate
		do
		end

feature {NONE} -- Assertions
	state_printing_function_name: ET_FEATURE_NAME is
			-- Name of `state_printing_function' for this program point
		do
			create {ET_IDENTIFIER} Result.make (surrounding_class.lower_name + "_" +
				identifier_name (surrounding_routine.name).lower_name + "_surrogate_loop" + position.out + "_inv_")
		end

feature {NONE} -- Implementation
	is_modified (a_variable: CI_VARIABLE): BOOLEAN is
			-- Does `a_variable' correspond to one of the entities stored in `modified_variables'?
		do
			from
				modified_variables.start
			until
				modified_variables.after or Result
			loop
				if a_variable.is_same_writable (modified_variables.item) then
					Result := True
				end
				modified_variables.forth
			end
		end

	prune_modified_nodes (a_tree: TWO_WAY_TREE [CI_VARIABLE]) is
			-- Prune nodes from `a_tree' (together with their subtrees) if they correspond to one of the entities stored in `modified_variables'
		do
			from
				a_tree.child_start
			until
				a_tree.child_after
			loop
				if is_modified (a_tree.child_item) then
					a_tree.remove_child
				else
					prune_modified_nodes (a_tree.child)
					a_tree.child_forth
				end
			end
		end
invariant
	loop_point_exists: loop_point /= Void
	modified_variables_exists: modified_variables /= Void
end
